#include <stdio.h>
#include "xtdio.h"

/* ---------------------------------------------------------------------
 * Definitions:
 */
#define BUFINCR	64

/* --------------------------------------------------------------------- */

/* mkbkupfile: renames a filename to a backup filename, and replaces
 *             the filename-file with the repfile-filename
 *             In summary: repfile -> filename -> bakfile
 *
 *  Returns 0=success  1=failure
 */
int mkbkupfile(char *filename,char *repfile)
{
size_t        flen;
static char  *bakfile= NULL;
static size_t baklen = 0;


/* set up bakfile storage */
flen= strlen(filename) + 4;
if(flen >= baklen) {
	baklen= flen + BUFINCR;
	if(!bakfile) bakfile= (char *) calloc(baklen,sizeof(char));
	else         bakfile= (char *) realloc(bakfile,baklen);
	if(!bakfile) {
		error(XTDIO_WARNING,"mkbkupfile: unable to allocate bakfile storage (%u bytes)\n",baklen);
		return 1;
		}
	}

/* create backup filename */

#if defined(vms)
	{
	/* keep old suffix separated with underscore and append .bak */
	char *dot;
	char *rb;
	dot= strrchr(filename,'.');
	rb = strrchr(filename,']');
	if(dot < rb) dot= NULL;
	if(dot) *dot= '_';
	sprintf(bakfile,"%s.bak",filename);
	if(dot) *dot= '.'; /* restore dot */
	}

#elif defined(msdos)
	{
	/* lose old suffix and append .bak */
	char *dot;
	dot= strrchr(filename,'.');
	if(dot) *dot= '\0';
	sprintf(bakfile,"%s.bak",filename);
	if(dot) *dot= '.'; /* restore dot */
	}

#else
/* keep original filename but append .bak */
sprintf(bakfile,"%s.bak",filename);

#endif


/* rename old filename file to the bakfile */
if(rename(filename,bakfile) == -1) {
	remove(bakfile);
	if(rename(filename,bakfile) == -1) {
		error(XTDIO_WARNING,"unable to rename <%s%s%s> to <%s%s%s>\n",
		  RVGREEN,filename,GREEN,
		  RVGREEN,bakfile,GREEN);
		return 1;
		}
	}

/* rename the (new) repfile to filename */
if(rename(repfile,filename) == -1) {
	error(XTDIO_WARNING,"unable to rename <%s%s%s> to <%s%s%s>\n",
	  RVGREEN,repfile,GREEN,
	  RVGREEN,filename,GREEN);
	return 1;
	}

return 0;
}

/* --------------------------------------------------------------------- */
